import 'dart:async';
import 'dart:html'; // ignore: avoid_web_libraries_in_flutter
import 'dart:typed_data';

import 'package:cross_file/cross_file.dart';
import 'package:cross_file/src/web_helpers/web_helpers.dart'; // ignore: implementation_imports
import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_html/shims/dart_ui.dart' as ui;
import 'package:flutterx_drop_view/src/drag_operation.dart';
import 'package:flutterx_drop_view/src/mime_type_not_match_exception.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_utils/flutterx_utils.dart';
import 'package:mime/mime.dart';

class RawDropView extends StatefulWidget {
  final DragOperation dragOperation;
  final SystemMouseCursor? cursor;
  final Set<String> mimeTypes;
  final bool throwOnMimeTypeNotMatch;
  final bool multiple;
  final void Function(Stream<XFile> files) onDrop;
  final Widget Function(BuildContext context, LiveData<bool> dragging) builder;

  const RawDropView({
    Key? key,
    this.dragOperation = DragOperation.copy,
    this.cursor,
    this.mimeTypes = const {},
    this.throwOnMimeTypeNotMatch = false,
    this.multiple = false,
    required this.onDrop,
    required this.builder,
  }) : super(key: key);

  @override
  State<RawDropView> createState() => _RawDropViewState();

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(EnumProperty<DragOperation>('dragOperation', dragOperation))
      ..add(DiagnosticsProperty<SystemMouseCursor?>('cursor', cursor))
      ..add(IterableProperty<String>('mimeTypes', mimeTypes))
      ..add(DiagnosticsProperty<bool>('throwOnMimeTypeNotMatch', throwOnMimeTypeNotMatch))
      ..add(DiagnosticsProperty<bool>('multiple', multiple))
      ..add(ObjectFlagProperty<void Function(Stream<XFile> files)>.has('onDrop', onDrop))
      ..add(ObjectFlagProperty<Widget Function(BuildContext context, LiveData<bool> dragging)>.has('builder', builder));
  }

  static Stream<XFile> pickFiles({Set<String> mimeTypes = const {}, bool multiple = false}) async* {
    final picker = FileUploadInputElement();
    final safari = isSafari();
    if (safari) document.body!.append(picker);
    picker.multiple = multiple;
    if (mimeTypes.isNotEmpty && !mimeTypes.contains('*/*')) picker.accept = mimeTypes.join(',');
    picker.click();
    await picker.onChange.first;
    final files = picker.files ?? const [];
    if (safari) picker.remove();
    for (final file in files) yield await _readFromFile(file, lookupMimeType(file.name));
  }

  static Future<XFile> _readFromFile(File file, String? mimeType) {
    final reader = FileReader()..readAsArrayBuffer(file);
    return reader.onLoadEnd.first.then((_) {
      final bytes = reader.result as Uint8List?;
      if (bytes == null) throw Exception('Cannot read bytes from Blob. Is it still available?');
      return XFile.fromData(bytes,
          mimeType: mimeType, name: file.name, length: bytes.length, lastModified: file.lastModifiedDate);
    });
  }
}

class _RawDropViewState extends State<RawDropView> {
  final MutableLiveData<bool> _dragging = MutableLiveData(initialValue: false);
  late String _viewType;
  late DivElement _element;

  @override
  void initState() {
    super.initState();
    ui.platformViewRegistry.registerViewFactory(
        _viewType = 'drop-view-$hashCode',
        (viewId) => _element = DivElement()
          ..id = 'dropzone-$viewId'
          ..style.pointerEvents = 'auto'
          ..style.border = 'none'
          ..style.cursor = widget.cursor?.kind.snakeCase.replaceAll('_', '-')
          ..style.width = '100%'
          ..style.height = '100%'
          ..addEventListener('dragover', _onDragOver)
          ..addEventListener('dragleave', _onDragLeave)
          ..addEventListener('drop', _onDrop));
  }

  @override
  void didUpdateWidget(covariant RawDropView oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (oldWidget.cursor != widget.cursor) _element.style.cursor = widget.cursor?.kind.snakeCase.replaceAll('_', '-');
  }

  void _onDragOver(Event event) {
    if (event is! MouseEvent) return;
    event.preventDefault();
    event.dataTransfer.dropEffect = widget.dragOperation.name;
    _dragging.value = true;
  }

  void _onDragLeave(Event event) {
    if (event is! MouseEvent) return;
    event.preventDefault();
    _dragging.value = false;
  }

  void _onDrop(Event event) {
    if (event is! MouseEvent) return;
    event.preventDefault();
    widget.onDrop(_readFilesAsStream(event));
    _dragging.value = false;
  }

  @override
  Widget build(BuildContext context) => Stack(
      fit: StackFit.passthrough,
      alignment: Alignment.center,
      children: [HtmlElementView(viewType: _viewType), widget.builder(context, _dragging)]);

  Stream<XFile> _readFilesAsStream(MouseEvent event) async* {
    final files = _eventFiles(event);
    if (files.isEmpty) return;
    final mimeTypes = widget.mimeTypes;
    final checkMimeType = mimeTypes.isNotEmpty && !mimeTypes.contains('*/*');
    for (final file in widget.multiple ? files : files.skip(files.length - 1)) {
      final mimeType = lookupMimeType(file.name);
      if (checkMimeType && (mimeType == null || !mimeTypes.contains(mimeType))) {
        if (widget.throwOnMimeTypeNotMatch)
          throw MimeTypeNotMatchException(mimeTypes, mimeType);
        else
          continue;
      }
      yield await RawDropView._readFromFile(file, mimeType);
    }
  }

  static Iterable<File> _eventFiles(MouseEvent event) sync* {
    final files = event.dataTransfer.files;
    if (files != null) {
      yield* files;
    } else {
      final items = event.dataTransfer.items;
      if (items != null) {
        for (var i = 0; i < (items.length ?? 0); i++) {
          final item = items[i];
          if (item.kind != 'file') throw StateError('Invalid transfer item kind: ${item.kind}, expected file');
          final file = item.getAsFile();
          if (file == null) throw Exception('Cannot read transfer item as file. Is it still available?');
          yield file;
        }
      }
    }
  }
}
